;;;; crunch - syntax for aggregate types for use in crunched code

(module (crunch aggregate-types)

  (define-enum
   define-struct
   define-union
   define-compound-accessors)

  (import (scheme base)
  		  (chicken syntax)
  		  (crunch c))

  (import-for-syntax (crunch compiler type)
  					 (chicken plist)
  					 matchable)

  (define-syntax define-enum
    (er-macro-transformer
      (lambda (x r c)
        (let ((ds (r 'define-syntax))
              (sr (r 'syntax-rules))
              (se (r 'syntax-error))
              (id (r (cadr x)))
              (name (r 'name))
              (comps (caddr x)))
          `(,ds ,id
                (,sr ,(map (lambda (c) (if (pair? c) (car c) c)) comps)
                     ,@(let rec ((comps comps) (i 0))
                         (match comps
                           (()
                             `(((_ name) (,se "invalid enumeration value" ,id name))))
                           (((cname n) more ...)
                             `(((_ ,cname) ,n) ,@(rec more (+ n 1))))
                           ((cname more ...)
                             `(((_ ,cname) ,i) ,@(rec more (+ i 1))))))))))))

  (define-syntax define-struct
    (syntax-rules (typename)
      ((_ name slots ...)
        (define-struct-or-union struct name slots ...))))

  (define-syntax define-union
    (syntax-rules (typename)
      ((_ name slots ...)
        (define-struct-or-union union name slots ...))))

  (define-syntax define-struct-or-union
    (er-macro-transformer
      (lambda (x r c)
        (let* ((type (strip-syntax (cadr x)))
               (name (strip-syntax (caddr x)))
               (sname (symbol->string name))
               (slots (cdddr x))
               (%c-declare (r 'c-declare))
               (info (list type))
               (body
                 `(,%c-declare
                    ,(string-append
                       (if (eq? type 'typename)
                       	   ""
                       	   (symbol->string type))
                       " " sname " {\n"
                       (let rec ((slots slots))
                         (match slots
                           (() "};")
                           ((((s (? integer? n)) t) slots ...)
                             (set! info (cons (list s t n) info))
                             (string-append
                               (typevar->c-type (typespec t))
                               " "
                               (symbol->string (strip-syntax s))
                               "[ " (number->string n) " ];\n"
                               (rec slots)))
                           (((s t) slots ...)
                             (set! info (cons (list s t) info))
                             (string-append
                               (typevar->c-type (typespec t))
                               " "
                               (symbol->string (strip-syntax s))
                               ";\n"
                               (rec slots)))
                           (s (error "malformed slot specification"
                                 (strip-syntax s)))))))))
           (put! name 'crunch:struct (reverse info))
           body))))

  (define-syntax declare-struct
    (syntax-rules (typename)
      ((_ (typename name) slots ...)
        (declare-struct-or-union typename name slots ...))
      ((_ name slots ...)
        (declare-struct-or-union struct name slots ...))))

  (define-syntax declare-union
    (syntax-rules (typename)
      ((_ (typename name) slots ...)
        (declare-struct-or-union typename name slots ...))
      ((_ name slots ...)
        (declare-struct-or-union union name slots ...))))

  (define-syntax declare-struct-or-union
    (er-macro-transformer
      (lambda (x r c)
        (let* ((type (strip-syntax (cadr x)))
               (name (strip-syntax (caddr x)))
               (slots (cdddr x))
               (info (list type)))
          (let loop ((slots slots))
            (match slots
              (()
               (put! name 'crunch:struct (reverse info))
               '(##core#undefined))
              ((((s (? integer? n)) t) slots ...)
               (set! info (cons (list s t n) info))
               (loop slots)))
            (((s t) slots ...)
             (set! info (cons (list s t) info))
             (loop slots)))
          (s (error "malformed slot specification"
                    (strip-syntax s)))))))

  (define-syntax define-compound-accessors
    (er-macro-transformer
      (lambda (x r c)
        (let* ((%begin (r 'begin))
               (%define (r 'define))
               (%c-lambda (r 'c-lambda))
               (name #f)
               (info #f)
               (sname #f)
               (type #f))
          (define (getinfo t n)
            (set! type (strip-syntax t))
            (set! name (strip-syntax n))
            (set! info (get name 'crunch:struct))
            (unless info
              (error "struct or union not defined" name))
            (unless (eq? type (car info))
              (error "struct accessed as union or vice-versa" name))
            (set! sname (symbol->string name)))
          (define (%pointer? s) (c s 'pointer))
          (define (tprefix)
            (if (eq? type 'typename)
                ""
                (string-append (symbol->string type) " ")))
          (define (find-slot name)
            (let loop ((info (cdr info)))
              (match info
                (((n t . _) more ...)
                 (if (eq? name n)
                     (car info)
                     (loop more)))
                (_ (error "slot not found in constructor" name)))))
          (define (genvctor ctor args)
            (let ((cargs (map (lambda _ (gensym 'a)) args)))
              `(,%define
                 ,ctor
                 (,%c-lambda
                   ,(map (lambda (arg name)
                           (let ((t (cadr (find-slot (strip-syntax arg)))))
                             (list t name)))
                      args cargs)
                   (,(car info) ,name)
                   ,(apply
                      string-append
                      (tprefix) sname " _s;\n"
                      (append
                        (map (lambda (arg name)
                               (string-append
                                 "_s." (symbol->string (strip-syntax arg))
                                 " = " (symbol->string name) ";\n"))
                          args cargs)
                        '("return _s;")))))))
          (define (genpctor ctor args)
            (let ((cargs (map (lambda _ (gensym 'a)) args)))
              `(,%define
                 ,ctor
                 (,%c-lambda
                   ,(map (lambda (arg name)
                           (let ((t (cadr (find-slot (strip-syntax arg)))))
                             (list t name)))
                      args cargs)
                   (pointer (,(car info) ,name))
                   ,(apply
                      string-append
                      (tprefix) sname
                      " *_s = malloc(sizeof(" (symbol->string (car info))
                      " " sname "));\n"
                      (append
                        (map (lambda (arg name)
                               (string-append
                                 "_s->" (symbol->string (strip-syntax arg))
                                 " = " (symbol->string name) ";\n"))
                          args cargs)
                        '("return _s;")))))))
          (define (genvslots slots)
            (define (setter s set)
              (match (find-slot (strip-syntax s))
                ((s t n)
                 `(,%define ,set
                            (,%c-lambda
                              (((,type ,name) _s) (integer _i) (,t _x))
                              void
                              ,(string-append
                                 "_s." (symbol->string s)
                                 "[ _i ] = _x;"))))
                ((s t)
                 `(,%define ,set
                            (,%c-lambda
                              (((,type ,name) _s) (,t _x))
                              void
                              ,(string-append
                                 "_s." (symbol->string s)
                                 " = _x;"))))))
            (define (getter s get)
              (match (find-slot (strip-syntax s))
                ((s t n)
                 `(,%define ,get
                            (,%c-lambda
                              (((,type ,name) _s) (integer _i))
                              ,t
                              ,(string-append
                                 "return _s." (symbol->string s)
                                 "[ _i ];"))))
                ((s t)
                 `(,%define ,get
                            (,%c-lambda
                              (((,type ,name) _s))
                              ,t
                              ,(string-append
                                 "return _s." (symbol->string s) ";"))))))
            `(,%begin
               ,@(map (lambda (slot)
                        (match slot
                          ((s get set)
                           `(,%begin
                              ,(getter s get)
                              ,(setter s set)))
                          ((s get)
                           (getter s get))
                          (_ (error "malformed slot specification"
                                    (strip-syntax slot)))))
                   slots)))
          (define (genpslots slots)
            (define (setter s set)
              (match (find-slot (strip-syntax s))
                ((s t n)
                 `(,%define ,set
                            (,%c-lambda
                              (((pointer (,type ,name)) _s) (integer _i) (,t _x))
                              void
                              ,(string-append
                                 "_s->" (symbol->string s)
                                 "[ _i ] = _x;"))))
                ((s t)
                 `(,%define ,set
                            (,%c-lambda
                              (((pointer (,type ,name)) _s) (,t _x))
                              void
                              ,(string-append
                                 "_s->" (symbol->string s)
                                 " = _x;"))))))
            (define (getter s get)
              (match (find-slot (strip-syntax s))
                ((s t n)
                 `(,%define ,get
                            (,%c-lambda
                              (((pointer (,type ,name)) _s) (integer _i))
                              ,t
                              ,(string-append
                                 "return _s->" (symbol->string s)
                                 "[ _i ];"))))
                ((s t)
                 `(,%define ,get
                            (,%c-lambda
                              (((pointer (,type ,name)) _s))
                              ,t
                              ,(string-append
                                 "return _s->" (symbol->string s) ";"))))))
            `(,%begin
               ,@(map (lambda (slot)
                        (match slot
                          ((s get set)
                           `(,%begin
                              ,(getter s get)
                              ,(setter s set)))
                          ((s get)
                           (getter s get))
                          (_ (error "malformed slot specification"
                                    (strip-syntax slot)))))
                   slots)))
          (match x
            ((_ ((? %pointer?) (type name)) (ctor args ...) slots ...)
             (getinfo type name)
             `(,%begin
                ,(genpctor ctor args)
                ,(genpslots slots)))
            ((_ ((? %pointer?) (? symbol? name)) (ctor args ...) slots ...)
             (getinfo 'typename name)
             `(,%begin
                ,(genpctor ctor args)
                ,(genpslots slots)))
            ((_ (type name) (ctor args ...) slots ...)
             (getinfo type name)
             `(,%begin
                ,(genvctor ctor args)
                ,(genvslots slots)))
            ((_ (? symbol? name) (ctor args ...) slots ...)
             (getinfo 'typename name)
             `(,%begin
                ,(genvctor ctor args)
                ,(genvslots slots)))
            (_ (error "malformed record-type definition" (strip-syntax x))))))))

)